#include <QVBoxLayout>
#include <QFile>

#include "MainWindow.h"
#include "ConsoleView.h"
#include "Emulator.h"

#define MESSAGE_UNKNOWN_COMMAND "  Unknown command.\n"
#define MESSAGE_WRONG_VALUE     "  Wrong value.\n"

ConsoleView::ConsoleView( MainWindow *mainWindow0, QWidget *parent ):
	QWidget( parent ) {

  mainWindow = mainWindow0;
  m_okCurrentProc = FALSE;

  consoleTE = new QTextEdit();
  consoleTE->setReadOnly( true );
  consoleTE->document()->setMaximumBlockCount( 200 );
  consoleTE->setFocusPolicy( Qt::NoFocus );

  consoleLE = new QLineEdit();
  consoleLE->setFocus();

  QVBoxLayout *layout = new QVBoxLayout();
  layout->addWidget( consoleTE );
  layout->addWidget( consoleLE );

  setLayout( layout );

  connect( consoleLE, SIGNAL(returnPressed()), SLOT(DoConsoleCommand()) );

  setMinimumSize( 200, 100 );
  show();

  Print( "UKNC Everywhere!\n\n"
         "Debug console. Type 'h' to show help.\n\n" );
  PrintConsolePrompt();
}

CProcessor* ConsoleView::GetCurrentProcessor() {
    if (m_okCurrentProc)
        return g_pBoard->GetCPU();
    else
        return g_pBoard->GetPPU();
}

void ConsoleView::Print( const QString& message ) {
  QTextCursor cursor = consoleTE->textCursor();
  cursor.movePosition( QTextCursor::End );
  cursor.insertText( message );
  consoleTE->setTextCursor( cursor );
}

void ConsoleView::ClearConsole() {
  consoleTE->clear();
}

void ConsoleView::PrintConsolePrompt() {
    CProcessor* pProc = GetCurrentProcessor();
    TCHAR bufferAddr[7];
    PrintOctalValue(bufferAddr, pProc->GetPC());
    Print( QString("%1:%2> ").arg(pProc->GetName()).arg(bufferAddr) );
}

void ConsoleView::PrintRegister( QString strName, WORD value ) {
  TCHAR bufferOct[7], bufferBin[17];
  PrintOctalValue(bufferOct, value);
  PrintBinaryValue(bufferBin, value);
  Print( QString("  %1 %2 %3\n").arg(strName).arg(bufferOct).arg(bufferBin) );
}

void ConsoleView::SaveMemoryDump( CProcessor *pProc ) {
        CMemoryController* pMemCtl = pProc->GetMemoryController();
        BYTE buf[65536];

        for ( int i = 0; i < 65536; i++ )
                buf[i] = pMemCtl->GetByte(i, 1);

  QFile file( QString("memdump%1.bin").arg(pProc->GetName()) );
  if ( file.open( QFile::WriteOnly ) )
	file.write( (char*)buf, sizeof(buf) );
}

void ConsoleView::PrintMemoryDump( CProcessor* pProc, WORD address, int lines ) {
  address &= ~1;  // Line up to even address
  
  CMemoryController* pMemCtl = pProc->GetMemoryController();
  BOOL okHaltMode = pProc->IsHaltMode();

  for (int line = 0; line < lines; line++) {
	WORD dump[8];
	for (int i = 0; i < 8; i++)
            dump[i] = pMemCtl->GetWord(address + i*2, okHaltMode);
	
        TCHAR pBuf[7];
        PrintOctalValue(pBuf, address);

	QString buffer = QString("  %1  ").arg( pBuf );

        for (int i = 0; i < 8; i++) {
            PrintOctalValue(pBuf, dump[i]);
            buffer += QString("%1 ").arg( pBuf );
        }
        buffer += " ";

        for (int i = 0; i < 8; i++) {
            WORD word = dump[i];

            BYTE ch1 = LOBYTE(word);
            unsigned short wch1 = Translate_KOI8R_Unicode(ch1);
            if (ch1 < 32) wch1 = '.';
            buffer += QChar( wch1 );

            BYTE ch2 = HIBYTE(word);
            unsigned short wch2 = Translate_KOI8R_Unicode(ch2);
            if (ch2 < 32) wch2 = _T('.');
            buffer += QChar( wch2 );
        }

	Print( buffer + "\n" );

        address += 16;
    }
}

void ConsoleView::PrintDisassemble(CProcessor* pProc, WORD address, BOOL okOneInstr) {
    CMemoryController* pMemCtl = pProc->GetMemoryController();
    BOOL okHaltMode = pProc->IsHaltMode();

    const int nWindowSize = 30;
    WORD memory[nWindowSize + 2];
    BOOL okValid;
    for (int i = 0; i < nWindowSize + 2; i++)
        memory[i] = pMemCtl->GetWordView(address + i*2, okHaltMode, TRUE, &okValid);

    TCHAR bufaddr[7];
    TCHAR bufvalue[7];

    int length = 0;
    for (int index = 0; index < nWindowSize; index++) {  //  
        PrintOctalValue(bufaddr, address);
        WORD value = memory[index];
        PrintOctalValue(bufvalue, value);

        if (length > 0)
            Print( QString("  %1  %2\n").arg(bufaddr).arg(bufvalue) );
        else {
            if (okOneInstr && index > 0) break;

            TCHAR instr[8];
            TCHAR args[32];
            length = DisassembleInstruction(memory + index, address, instr, args);
            if (index + length > nWindowSize) break;

            Print( QString("  %1  %2  %3 %4\n").arg(bufaddr).arg(bufvalue)
		.arg( QString(instr).leftJustified(7) ).arg(args) );
        }
        length--;
        address += 2;
    }
}

void ConsoleView::Step() {
  consoleLE->setText( "s" );
  DoConsoleCommand();
}

void ConsoleView::ShowHelp()
{   
    Print("Console command list:\n"
            "  c          Clear console log\n"
            "  dXXXXXX    Disassemble from address XXXXXX\n"
            "  g          Go; free run\n"
            "  gXXXXXX    Go; run processor until breakpoint at address XXXXXX\n"
            "  m          Memory dump at current address\n"
            "  mXXXXXX    Memory dump at address XXXXXX\n"
            "  mrN        Memory dump at address from register N; N=0..7\n"
            "  p          Switch to other processor\n"
            "  r          Show register values\n"
            "  rN         Show value of register N; N=0..7,ps\n"
            "  rN XXXXXX  Set register N to value XXXXXX; N=0..7,ps\n"
            "  s          Step; executes one instruction (F8)\n"
            "  u          Save memory dump to file memdumpXPU.bin\n"
        );
}

void ConsoleView::DoConsoleCommand() {
    QByteArray commandArr = consoleLE->text().toLocal8Bit();
    const char *command = commandArr.data();

    consoleLE->clear();

    if ( commandArr.isEmpty() ) return;  // Nothing to do

    // Echo command to the log
    Print( commandArr + "\n" );

    BOOL okUpdateAllViews = FALSE;  // Flag - need to update all debug views
    CProcessor* pProc = GetCurrentProcessor();

    // Execute the command
    switch ( command[0] ) {
    case 'h':
        ShowHelp();
        break;
    case 'c':  // Clear log
        ClearConsole();
        break;
    case 'p':  // Switch CPU/PPU 
        m_okCurrentProc = !m_okCurrentProc;
        mainWindow->debugView()->setCurrentProc(m_okCurrentProc);   // Switch DebugView to current processor
        break;
    case 'r':  // Register operations
        if (command[1] == 0) { // Print all registers
            for (int r = 0; r < 8; r++)
                PrintRegister( REGISTER_NAME[r], pProc->GetReg(r) );
	}
        else if (command[1] >= '0' && command[1] <= '7') { // "r0".."r7"
            int r = command[1] - '0';
            if (command[2] == 0) // "rN" - show register N
                PrintRegister( REGISTER_NAME[r], pProc->GetReg(r) );
            else if (command[2] == '=' || command[2] == ' ') {  // "rN=XXXXXX" - set register N to value XXXXXX
                WORD value;
                if ( !ParseOctalValue(command + 3, &value) )
                    Print(MESSAGE_WRONG_VALUE);
                else {
                    pProc->SetReg(r, value);
                    PrintRegister(REGISTER_NAME[r], value);
                    okUpdateAllViews = TRUE;
                }
            }
            else
                Print(MESSAGE_UNKNOWN_COMMAND);
        }
        else if (command[1] == 'p' && command[2] == 's') { // "rps"
            if (command[3] == 0)  // "rps" - show PSW
                PrintRegister("PS", pProc->GetPSW());
            else if (command[3] == '=' || command[3] == ' ') {  // "rps=XXXXXX" - set PSW to value XXXXXX
                WORD value;
                if (! ParseOctalValue(command + 4, &value))
                    Print(MESSAGE_WRONG_VALUE);
                else  {
                    pProc->SetPSW(value);
                    PrintRegister("PS", value);
                    okUpdateAllViews = TRUE;
                }
            }
            else
                Print(MESSAGE_UNKNOWN_COMMAND);
        }
        else
            Print(MESSAGE_UNKNOWN_COMMAND);
        break;
    case 's':  // Step - execute one instruction
        PrintDisassemble(pProc, pProc->GetPC(), TRUE);
        //pProc->Execute();
                
                g_pBoard->DebugTicks();

        okUpdateAllViews = TRUE;
        break;
    case 'd':  // Disassemble
        if (command[1] == 0)  // "d" - disassemble at current address
            PrintDisassemble(pProc, pProc->GetPC(), FALSE);
        else if (command[1] >= '0' && command[1] <= '7') {  // "dXXXXXX" - disassemble at address XXXXXX
            WORD value;
            if ( !ParseOctalValue(command + 1, &value) )
                Print(MESSAGE_WRONG_VALUE);
            else
                PrintDisassemble(pProc, value, FALSE);
        }
        else
            Print(MESSAGE_UNKNOWN_COMMAND);
        break;
    case 'u':
        SaveMemoryDump(pProc);
        break;
    case 'm':
        if (command[1] == 0)  // "m" - dump memory at current address
            PrintMemoryDump(pProc, pProc->GetPC(), 8);
        else if (command[1] >= '0' && command[1] <= '7') { // "mXXXXXX" - dump memory at address XXXXXX
            WORD value;
            if ( !ParseOctalValue(command + 1, &value) )
                Print(MESSAGE_WRONG_VALUE);
            else
                PrintMemoryDump(pProc, value, 8);
        }
        else if (command[1] == 'r' &&
                command[2] >= '0' && command[2] <= '7') {  // "mrN" - dump memory at address from register N
            WORD address = pProc->GetReg( command[2] - '0' );
            PrintMemoryDump(pProc, address, 8);
        }
        else
            Print(MESSAGE_UNKNOWN_COMMAND);
        break;
    //TODO: "mXXXXXX YYYYYY" - set memory cell at XXXXXX to value YYYYYY
    //TODO: "mrN YYYYYY" - set memory cell at address from rN to value YYYYYY
    case 'g':
        if (command[1] == 0)
            Emulator_Start();
        else {
            WORD value;
            if ( !ParseOctalValue(command + 1, &value) )
                Print(MESSAGE_WRONG_VALUE);
            else {
                if (m_okCurrentProc)
                    Emulator_SetCPUBreakpoint(value);
                else
                    Emulator_SetPPUBreakpoint(value);
                Emulator_Start();
            }
        }
        break;
    default:
        Print(MESSAGE_UNKNOWN_COMMAND);
        break;
    }

    PrintConsolePrompt();
    
    if (okUpdateAllViews)
        mainWindow->updateAllViews();
}
