#include <QEvent>
#include <QMenu>
#include <QMouseEvent>
#include <QLineEdit>
#include <QTimer>
#include "DebugView.h"
#include "Common.h"
#include "Emulator.h"

// Colors
#define COLOR_RED   QColor(255,0,0)
#define COLOR_BLUE  QColor(0,0,255)

DebugView::DebugView( QWidget *parent ): QWidget(parent) {
  setContextMenuPolicy( Qt::CustomContextMenu );
  connect( this, SIGNAL(customContextMenuRequested(const QPoint&)),
        SLOT(showMenu()) );

  setMouseTracking( true );
  objects.resize( 150 );
  selectedObj = -1;
  valueEditor = 0;

  m_okDebugProcessor = FALSE;
  updateFontMetrics();
  show();
}

void DebugView::showMenu() {
  QMenu *menu = new QMenu( this );

  QActionGroup *group = new QActionGroup( menu );

  QAction *action = menu->addAction( "CPU", this, SLOT(debugCPU()) );
  group->addAction( action );
  action->setCheckable( true );
  action->setChecked( m_okDebugProcessor );

  action = menu->addAction( "PPU", this, SLOT(debugPPU()) );
  group->addAction( action );
  action->setCheckable( true );
  action->setChecked( !m_okDebugProcessor );

  menu->popup( QCursor::pos() );
}

void DebugView::debugCPU() {
  m_okDebugProcessor = true;
  repaint();
}

void DebugView::debugPPU() {
  m_okDebugProcessor = false;
  repaint();
}

bool DebugView::event( QEvent *ev ) {
  if ( ev->type() == QEvent::FontChange ) updateFontMetrics();
  return QWidget::event( ev );
}

void DebugView::updateFontMetrics() {
  QFontMetrics fm( font() );
  cxChar = fm.averageCharWidth();
  cyLine = fm.height();
  ascent = fm.ascent();

  resize( cxChar * 105, cyLine * 48 );
}

void DebugView::leaveEvent( QEvent* /*ev*/ ) {
  if ( selectedObj != -1 ) {
	selectedObj = -1;
	repaint();
  }
}

void DebugView::mouseMoveEvent( QMouseEvent *ev ) {
  QWidget::mouseMoveEvent( ev );

  bool changed = false;
  if ( selectedObj != -1 ) {
	if ( objects[selectedObj].contains( ev->pos() ) ) return;

	selectedObj = -1;
	changed = true;
  }

  for ( int i = 0; i < objects.size(); i++ )
	if ( !objects[i].isEmpty() && objects[i].contains( ev->pos() ) ) {
		selectedObj = i;
		changed = true;
		break;
	}

  if ( changed ) repaint();
}

void DebugView::mouseReleaseEvent( QMouseEvent *ev ) {
  if ( ev->button() != Qt::LeftButton ) return;

  if ( selectedObj != -1 && objects[selectedObj].slot ) {
	if ( valueEditor ) delete valueEditor;

	valueEditor = new QLineEdit( objects[selectedObj].text, this );
	valueEditor->setFrame( false );

	QRect r = objects[selectedObj];
	r.adjust( -3, 0, 2, 0 );

	valueEditor->setGeometry( r );

	valueEditor->show();
	valueEditor->setFocus();

	connect( valueEditor, SIGNAL(returnPressed()), SLOT(changeValue()) );
	valueEditor->installEventFilter( this );

	ev->accept();
  }
}

void DebugView::changeValue() {
  QMetaObject::invokeMethod( this, objects[selectedObj].slot,
	Q_ARG(QVariant, objects[selectedObj].parameter),
	Q_ARG(QString, valueEditor->text()) );

  valueEditor->deleteLater();
  valueEditor = 0;
}

bool DebugView::eventFilter( QObject *obj, QEvent *ev ) {
  if ( obj == valueEditor && ev->type() == QEvent::KeyPress ) {
	QKeyEvent *kev = (QKeyEvent*)ev;
	if ( kev->key() == Qt::Key_Escape ) {
		valueEditor->deleteLater();
		valueEditor = 0;
		return true;
	}
  }

  return false;
}

void DebugView::onUpdate() {
    CProcessor* pCPU = g_pBoard->GetCPU();
    ASSERT(pCPU != NULL);

    // Get new register values and set change flags
    for (int r = 0; r < 8; r++) {
        WORD value = pCPU->GetReg(r);
        m_okDebugCpuRChanged[r] = (m_wDebugCpuR[r] != value);
        m_wDebugCpuR[r] = value;
    }
    WORD pswCPU = pCPU->GetPSW();
    m_okDebugCpuRChanged[8] = (m_wDebugCpuR[8] != pswCPU);
    m_wDebugCpuR[8] = pswCPU;

    CProcessor* pPPU = g_pBoard->GetPPU();
    ASSERT(pPPU != NULL);

    // Get new register values and set change flags
    for (int r = 0; r < 8; r++) {
        WORD value = pPPU->GetReg(r);
        m_okDebugPpuRChanged[r] = (m_wDebugPpuR[r] != value);
        m_wDebugPpuR[r] = value;
    }
    WORD pswPPU = pPPU->GetPSW();
    m_okDebugPpuRChanged[8] = (m_wDebugPpuR[8] != pswPPU);
    m_wDebugPpuR[8] = pswPPU;
}

void DebugView::setCurrentProc( BOOL okCPU ) {
    m_okDebugProcessor = okCPU;
    repaint();
}

void DebugView::paintEvent( QPaintEvent* /* e */ ) {
    QPainter hdc( this );
    objIdx = 0;

    CProcessor* pDebugPU = (m_okDebugProcessor) ? g_pBoard->GetCPU() : g_pBoard->GetPPU();
    ASSERT(pDebugPU != NULL);
    WORD* arrR = (m_okDebugProcessor) ? m_wDebugCpuR : m_wDebugPpuR;
    BOOL* arrRChanged = (m_okDebugProcessor) ? m_okDebugCpuRChanged : m_okDebugPpuRChanged;
    WORD prevPC = (m_okDebugProcessor) ? g_wEmulatorPrevCpuPC : g_wEmulatorPrevPpuPC;

    QString sProcName = pDebugPU->GetName();

    hdc.drawText(cxChar * 1, 2 * cyLine + ascent, sProcName.left(3) );
    drawProcessor(hdc, pDebugPU, cxChar * 6, 2 * cyLine + ascent, arrR, arrRChanged);

    // Draw disasseble and stack for the current processor
    drawDisassemble(hdc, pDebugPU, prevPC, 0, 18 * cyLine + ascent);
    drawMemoryForRegister(hdc, 6, pDebugPU, 35 * cxChar, 1 * cyLine + ascent);

    CMemoryController* pDebugMemCtl = pDebugPU->GetMemoryController();
    drawPorts(hdc, m_okDebugProcessor, pDebugMemCtl, g_pBoard, 57 * cxChar, 1 * cyLine + ascent);

    drawChannels(hdc, 75 * cxChar, 1 * cyLine + ascent);

    for ( int i = objIdx; i < objects.size(); i++ )
	objects[i] = DebugObject();

    if ( selectedObj != -1 ) {
	hdc.setBackground( QColor(0xff, 0xff, 0xe0) );
	hdc.setBackgroundMode( Qt::OpaqueMode );
	hdc.drawText( objects[selectedObj].x(), objects[selectedObj].y() + ascent, objects[selectedObj].text );
    }
}

void DebugView::DrawOctalValue( QPainter& p, int x, int y, WORD value ) {
    if ( objIdx < objects.size() ) {
	objects[objIdx] = DebugObject( x, y - ascent, cxChar * 6, cyLine,
		QString::number(value, 8).rightJustified(6, '0'),
		objects[objIdx].slot, objects[objIdx].parameter );
    	objIdx++;
    }
    ::DrawOctalValue( p, x, y, value );
}

void DebugView::DrawBinaryValue( QPainter& p, int x, int y, WORD value ) {
    if ( objIdx < objects.size() ) {
    	objects[objIdx] = DebugObject( x, y - ascent, cxChar * 16, cyLine,
    		QString::number(value, 2).rightJustified(16, '0'),
		objects[objIdx].slot, objects[objIdx].parameter );
    	objIdx++;
    }
    ::DrawBinaryValue( p, x, y, value );
}

void DebugView::setChangeHandler( const char *slot, QVariant p ) {
    if ( objIdx < objects.size() ) {
	objects[objIdx].slot = slot;
        objects[objIdx].parameter = p;
    }
}

void DebugView::drawProcessor( QPainter& hdc, CProcessor* pProc, int x, int y, WORD* arrR, BOOL* arrRChanged ) {
  hdc.setPen( Qt::black );

  QColor colorText = hdc.pen().color();
  hdc.drawRect( x - cxChar/2, y - 8 - cyLine + ascent, cxChar + 26 * cxChar, 8 + cyLine * 14 );

    // Registers
    for (int r = 0; r < 8; r++) {
        hdc.setPen( arrRChanged[r] ? COLOR_RED : colorText );

        LPCTSTR strRegName = REGISTER_NAME[r];
        hdc.drawText( x, y + r * cyLine, strRegName );

        WORD value = arrR[r]; //pProc->GetReg(r);

    	setChangeHandler( "setRegisterOctal", r );
        DrawOctalValue(hdc, x + cxChar * 3, y + r * cyLine, value);

	setChangeHandler( "setRegisterBinary", r );
        DrawBinaryValue(hdc, x + cxChar * 10, y + r * cyLine, value);
    }
    hdc.setPen( colorText );

    // CPC value
    hdc.drawText( x, y + 8 * cyLine, _T("PC'") );
    WORD cpc = pProc->GetCPC();
    DrawOctalValue(hdc, x + cxChar * 3, y + 8 * cyLine, cpc);
    DrawBinaryValue(hdc, x + cxChar * 10, y + 8 * cyLine, cpc);
    
    // PSW value
    hdc.setPen(arrRChanged[8] ? COLOR_RED : colorText);
    hdc.drawText( x, y + 10 * cyLine, _T("PS") );
    WORD psw = arrR[8]; // pProc->GetPSW();
    DrawOctalValue(hdc, x + cxChar * 3, y + 10 * cyLine, psw);
    hdc.drawText( x + cxChar * 10, y + 9 * cyLine, _T("       HP  TNZVC") );
    DrawBinaryValue(hdc, x + cxChar * 10, y + 10 * cyLine, psw);
   
    hdc.setPen(colorText);

    // CPSW value
    hdc.drawText( x, y + 11 * cyLine, _T("PS'") );
    WORD cpsw = pProc->GetCPSW();
    DrawOctalValue(hdc, x + cxChar * 3, y + 11 * cyLine, cpsw);
    DrawBinaryValue(hdc, x + cxChar * 10, y + 11 * cyLine, cpsw);

    // Processor mode - HALT or USER
    BOOL okHaltMode = pProc->IsHaltMode();
    hdc.drawText( x, y + 13 * cyLine, okHaltMode ? _T("HALT") : _T("USER") );

    // "Stopped" flag
    BOOL okStopped = pProc->IsStopped();
    if (okStopped)
        hdc.drawText( x + 6 * cxChar, y + 11 * cyLine, _T("STOP") );
}

void DebugView::setRegister( int regNum, int newValue ) {
  CProcessor* pDebugPU = (m_okDebugProcessor) ? g_pBoard->GetCPU() : g_pBoard->GetPPU();
  pDebugPU->SetReg( regNum, newValue );

  if (m_okDebugProcessor) m_wDebugCpuR[regNum] = newValue;
  else m_wDebugPpuR[regNum] = newValue;

  repaint();
}

void DebugView::setRegisterOctal( QVariant p, QString newValue ) {
  bool ok;
  WORD value = newValue.toInt( &ok, 8 );
  if ( ok ) setRegister( p.toInt(), value );
}

void DebugView::setRegisterBinary( QVariant p, QString newValue ) {
  bool ok;
  WORD value = newValue.toInt( &ok, 2 );
  if ( ok ) setRegister( p.toInt(), value );
}

void DebugView::drawDisassemble( QPainter& hdc, CProcessor* pProc, WORD previous, int x, int y ) {
  QColor colorText = hdc.pen().color();

    CMemoryController* pMemCtl = pProc->GetMemoryController();
    WORD current = pProc->GetPC();

    //      
    const int nWindowSize = 30;
    WORD memory[nWindowSize + 2];
    for (int idx = 0; idx < nWindowSize; idx++) {
        BOOL okValidAddress;
        memory[idx] = pMemCtl->GetWordView(
                current + idx * 2 - 10, pProc->IsHaltMode(), TRUE, &okValidAddress);
    }

    //TextOut(hdc, x, y, _T(" address  value   instruction"), 29);
    //y += cyLine;

    WORD address = current - 10;
    WORD disasmfrom = current;
    if (previous >= address && previous < current)
        disasmfrom = previous;

    int length = 0;
    for (int index = 0; index < nWindowSize; index++) {  //  
        ::DrawOctalValue(hdc, x + 5 * cxChar, y, address);  // Address
        // Value at the address
        WORD value = memory[index];

        setChangeHandler( "setMemoryValue", address );
        DrawOctalValue(hdc, x + 13 * cxChar, y, value);

        // Current position
        if (address == current)
            hdc.drawText( x + 1 * cxChar, y, _T("PC>>") );
        else if (address == previous)
        {
            hdc.setPen( COLOR_BLUE );
            hdc.drawText( x + 1 * cxChar, y, _T("  > ") );
        }

        if (address >= disasmfrom && length == 0) {
            TCHAR strInstr[8];
            TCHAR strArg[32];
            length = DisassembleInstruction(memory + index, address, strInstr, strArg);
            if (index + length <= nWindowSize)
            {
                hdc.drawText( x + 21 * cxChar, y, strInstr );
                hdc.drawText( x + 29 * cxChar, y, strArg );
            }
            hdc.setPen( colorText );
        }
        if (length > 0) length--;

        address += 2;
        y += cyLine;
    }
}

void DebugView::setMemoryValue( QVariant p, QString newValue ) {
  bool ok;
  WORD value = newValue.toInt( &ok, 8 );
  if ( ok ) {
  	CProcessor* pProc = (m_okDebugProcessor) ? g_pBoard->GetCPU() : g_pBoard->GetPPU();
	pProc->GetMemoryController()->SetWord(p.toInt(), pProc->IsHaltMode(), value);
	repaint();
  }
}

void DebugView::drawMemoryForRegister( QPainter& hdc, int reg, CProcessor* pProc, int x, int y ) {
    WORD current = pProc->GetReg(reg);
    BOOL okExec = (reg == 7);

    //      
    WORD memory[16];
    CMemoryController* pMemCtl = pProc->GetMemoryController();
    for (int idx = 0; idx < 16; idx++) {
        BOOL okValidAddress;
        memory[idx] = pMemCtl->GetWordView(
                current + idx * 2 - 16, pProc->IsHaltMode(), okExec, &okValidAddress);
    }

    WORD address = current - 16;
    for (int index = 0; index < 16; index++) {  //  
        // 
        ::DrawOctalValue(hdc, x + 4 * cxChar, y, address);

        //   
        WORD value = memory[index];

	setChangeHandler( "setMemoryValue", address );
        DrawOctalValue(hdc, x + 12 * cxChar, y, value);

        //  
        if (address == current) {
            hdc.drawText( x + 2 * cxChar, y, _T(">>") );
            hdc.drawText( x, y, REGISTER_NAME[reg] );
        }

        address += 2;
        y += cyLine;
    }
}

void DebugView::drawPorts( QPainter& hdc, BOOL okProcessor, CMemoryController* pMemCtl, CMotherboard* pBoard, int x, int y ) {
        hdc.drawText( x, y, _T("Ports:") );

    if (okProcessor)  // CPU
    {
        WORD value176640 = pMemCtl->GetPortView(0176640);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 1 * cyLine, 0176640);
	setChangeHandler( "setMemoryValue", 0176640 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 1 * cyLine, value176640);

        WORD value176642 = pMemCtl->GetPortView(0176642);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 2 * cyLine, 0176642);
	setChangeHandler( "setMemoryValue", 0176642 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 2 * cyLine, value176642);

        //TODO
    }
    else  // PPU
    {
        WORD value177010 = pMemCtl->GetPortView(0177010);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 1 * cyLine, 0177010);
	setChangeHandler( "setMemoryValue", 0177010 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 1 * cyLine, value177010);

        WORD value177012 = pMemCtl->GetPortView(0177012);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 2 * cyLine, 0177012);
	setChangeHandler( "setMemoryValue", 0177012 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 2 * cyLine, value177012);

        WORD value177014 = pMemCtl->GetPortView(0177014);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 3 * cyLine, 0177014);
	setChangeHandler( "setMemoryValue", 0177014 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 3 * cyLine, value177014);

        WORD value177016 = pMemCtl->GetPortView(0177016);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 4 * cyLine, 0177016);
	setChangeHandler( "setMemoryValue", 0177016 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 4 * cyLine, value177016);

        WORD value177020 = pMemCtl->GetPortView(0177020);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 5 * cyLine, 0177020);
	setChangeHandler( "setMemoryValue", 0177020 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 5 * cyLine, value177020);

        WORD value177022 = pMemCtl->GetPortView(0177022);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 6 * cyLine, 0177022);
	setChangeHandler( "setMemoryValue", 0177022 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 6 * cyLine, value177022);

        WORD value177024 = pMemCtl->GetPortView(0177024);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 7 * cyLine, 0177024);
	setChangeHandler( "setMemoryValue", 0177024 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 7 * cyLine, value177024);

        WORD value177026 = pMemCtl->GetPortView(0177026);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 8 * cyLine, 0177026);
	setChangeHandler( "setMemoryValue", 0177026 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 8 * cyLine, value177026);

        WORD value177054 = pMemCtl->GetPortView(0177054);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 9 * cyLine, 0177054);
	setChangeHandler( "setMemoryValue", 0177054 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 9 * cyLine, value177054);

        WORD value177700 = pMemCtl->GetPortView(0177700);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 10 * cyLine, 0177700);
	setChangeHandler( "setMemoryValue", 0177700 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 10 * cyLine, value177700);

        WORD value177716 = pMemCtl->GetPortView(0177716);
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 11 * cyLine, 0177716);
	setChangeHandler( "setMemoryValue", 0177716 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 11 * cyLine, value177716);

        WORD value177710 = pBoard->GetTimerStateView();
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 13 * cyLine, 0177710);
	setChangeHandler( "setMemoryValue", 0177710 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 13 * cyLine, value177710);

        WORD value177712 = pBoard->GetTimerReloadView();
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 14 * cyLine, 0177712);
	setChangeHandler( "setMemoryValue", 0177712 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 14 * cyLine, value177712);

        WORD value177714 = pBoard->GetTimerValueView();
        ::DrawOctalValue(hdc, x + 0 * cxChar, y + 15 * cyLine, 0177714);
	setChangeHandler( "setMemoryValue", 0177714 );
        DrawOctalValue(hdc, x + 8 * cxChar, y + 15 * cyLine, value177714);
    }
}

void DebugView::drawChannels( QPainter& hdc, int x, int y ) {
    hdc.drawText( x, y, _T("Channels:") );

    TCHAR bufData[7]; 
    TCHAR buffer[32];
        chan_stc tmpstc;


        tmpstc=g_pBoard->GetChannelStruct(0,0,0);
    
    PrintOctalValue(bufData, tmpstc.data);
    sprintf(buffer, _T("PPU CH:0 RX D:%s RDY:%d IRQ:%d"), bufData + 3, tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 1 * cyLine, buffer );
    
    tmpstc=g_pBoard->GetChannelStruct(0,1,0);
    PrintOctalValue(bufData, tmpstc.data);
    sprintf(buffer, _T("PPU CH:1 RX D:%s RDY:%d IRQ:%d"), bufData + 3, tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 2 * cyLine, buffer );

    tmpstc=g_pBoard->GetChannelStruct(0,2,0);
    PrintOctalValue(bufData, tmpstc.data);
    sprintf(buffer, _T("PPU CH:2 RX D:%s RDY:%d IRQ:%d"), bufData + 3, tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 3 * cyLine, buffer );

        tmpstc=g_pBoard->GetChannelStruct(0,0,1);
    sprintf(buffer, _T("PPU CH:0 TX       RDY:%d IRQ:%d"), tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 4 * cyLine, buffer );
    
        tmpstc=g_pBoard->GetChannelStruct(0,1,1);
    sprintf(buffer, _T("PPU CH:1 TX       RDY:%d IRQ:%d"), tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 5 * cyLine, buffer );


    tmpstc=g_pBoard->GetChannelStruct(1,0,0);
        PrintOctalValue(bufData, tmpstc.data);
    sprintf(buffer, _T("CPU CH:0 RX D:%s RDY:%d IRQ:%d"), bufData + 3, tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 6 * cyLine, buffer );

    tmpstc=g_pBoard->GetChannelStruct(1,1,0);
    PrintOctalValue(bufData, tmpstc.data);
    sprintf(buffer, _T("CPU CH:1 RX D:%s RDY:%d IRQ:%d"), bufData + 3, tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 7 * cyLine, buffer );

        tmpstc=g_pBoard->GetChannelStruct(1,0,1);
    sprintf(buffer, _T("CPU CH:0 TX       RDY:%d IRQ:%d"), tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 8 * cyLine, buffer );

        tmpstc=g_pBoard->GetChannelStruct(1,1,1);
    sprintf(buffer, _T("CPU CH:1 TX       RDY:%d IRQ:%d"), tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 9 * cyLine, buffer );

        tmpstc=g_pBoard->GetChannelStruct(1,2,1);
    sprintf(buffer, _T("CPU CH:2 TX       RDY:%d IRQ:%d"), tmpstc.ready, tmpstc.irq);
    hdc.drawText( x, y + 10 * cyLine, buffer );
}
