#ifndef MAINWINDOW_H_INCLUDED
#define MAINWINDOW_H_INCLUDED

#include <QMainWindow>

#include "Common.h"
#include "Emulator.h"
#include "UKNCBTL.h"
#include "ScreenView.h"
#include "KeyboardView.h"
#include "DebugView.h"
#include "MemoryView.h"
#include "ConsoleView.h"
#include "ui_MainWindow.h"
#include "ui_AboutDlg.h"

class MainWindow;

class CentralWidget: public QWidget {
public:
  CentralWidget( MainWindow *mainWindow, QWidget *parent = 0 ): QWidget(parent) {
    setFocusPolicy( Qt::NoFocus );

    screenView = new ScreenView( this );
    screenView->move( 0, 0 );

    keyboardView = new KeyboardView( screenView, this );

    consoleView  = new ConsoleView( mainWindow, this );
    consoleView->resize( screenView->width(), consoleView->height() );

    debugView    = new DebugView( this );

    memoryView   = new MemoryView( this );
    memoryView->resize( debugView->width(), memoryView->height() );

    adjustLayout();
    show();
  }

  void adjustLayout() {
    keyboardView->move( (screenView->width() - keyboardView->width()) / 2,
	screenView->height() + 8 );

    if ( keyboardView->isHidden() )
	consoleView->move( 0, screenView->height() );
    else
	consoleView->move( 0, keyboardView->y() + keyboardView->height() );

    debugView->move( screenView->width(), 0 );
    memoryView->move( screenView->width(), debugView->height() );

    consoleView->resize( consoleView->width(), consoleView->minimumHeight() );
    consoleView->resize( consoleView->width(),
	childrenRect().height() - consoleView->y() );

    resize( childrenRect().size() );
  }

  void setDebugViewFont( const QFont& font ) {
    consoleView->setFont( font );
    debugView->setFont( font );

    memoryView->setFont( font );
    memoryView->resize( debugView->width(), memoryView->height() );
  }

  void systemFrame() {
	screenView->prepareScreen();
	screenView->update();
  }

  ScreenView   *screenView;
  KeyboardView *keyboardView;
  DebugView    *debugView;
  MemoryView   *memoryView;
  ConsoleView  *consoleView;
};

class AboutDlg: public QDialog, public Ui_AboutDlg {
public:
  AboutDlg( QWidget *parent = 0 ): QDialog( parent ) {
	setupUi( this );
  }
};

class MainWindow: public QMainWindow, public Ui_MainWindow {
  Q_OBJECT
public:
  MainWindow();
  void restoreSettings();
  void updateMenu();
  ScreenView *screenView() { return cw->screenView; }
  DebugView *debugView() { return cw->debugView; }
  void updateAllViews() {
	Emulator_OnUpdate();
	cw->debugView->onUpdate();
	repaint();
	cw->debugView->repaint();
	cw->memoryView->repaint();
  }
  void systemFrame() { cw->systemFrame(); }

protected:
  void adjustLayout();

  CentralWidget *cw;
  void showHideDebug( bool show ) {
	cw->consoleView->setVisible( show );
  	cw->debugView->setVisible( show );
	cw->memoryView->setVisible( show );
	adjustLayout();
  }
  void showHideKeyboard( bool show ) {
	cw->keyboardView->setVisible( show );
	adjustLayout();
  }

  void doViewScreenMode( ScreenViewMode mode ) {
	Settings_SetScreenViewMode( mode );
	screenView()->setMode( mode );
  }
  void doViewHeightMode( int mode ) {
	Settings_SetScreenHeightMode( mode );
	screenView()->setHeightMode( mode );
	adjustLayout();
  }

  void doEmulatorFloppy( int slot );
  void doEmulatorCartridge( int slot );

protected slots:
  void doFileLoadState();
  void doFileSaveState();
  void doFileScreenshot();
  void doViewDebug( bool show ) {
  	Settings_SetDebug( show );
	showHideDebug( show );
  }
  void doViewKeyboard( bool show ) {
	Settings_SetKeyboard( show );
	showHideKeyboard( show );
  }
  void doViewScreenRGB() { doViewScreenMode( RGBScreen ); }
  void doViewScreenGRB() { doViewScreenMode( GRBScreen ); }
  void doViewScreenGray() { doViewScreenMode( GrayScreen ); }
  void doViewHeightNormal() { doViewHeightMode( 1 ); }
  void doViewHeightDouble() { doViewHeightMode( 2 ); }
  void selectDebugViewFont();
  void doEmulatorRun() {
	if ( g_okEmulatorRunning ) Emulator_Stop();
	else Emulator_Start();
  }
  void doStep() {
	cw->consoleView->Step();
  }
  void doEmulatorAutostart( bool on ) {
	Settings_SetAutostart( on );
  }
  void doEmulatorReset() { Emulator_Reset(); }
  void doEmulatorFloppy0() { doEmulatorFloppy( 0 ); }
  void doEmulatorFloppy1() { doEmulatorFloppy( 1 ); }
  void doEmulatorFloppy2() { doEmulatorFloppy( 2 ); }
  void doEmulatorFloppy3() { doEmulatorFloppy( 3 ); }
  void doEmulatorCartridge1() { doEmulatorCartridge( 1 ); }
  void doEmulatorCartridge2() { doEmulatorCartridge( 2 ); }
  void showAboutDlg() {
	AboutDlg d;
        d.exec();
  }
};

#endif // MAINWINDOW_H_INCLUDED
