#include <QPainter>
#include <QEvent>
#include <QMenu>
#include <QInputDialog>
#include <QLineEdit>
#include <QMouseEvent>

#include "MemoryView.h"
#include "Common.h"
#include "Emulator.h"
#include "emubase/Emubase.h"

MemoryView0::MemoryView0( QScrollBar *sb, QWidget *parent ): QWidget(parent),
	scrollBar(sb) {

  setContextMenuPolicy( Qt::CustomContextMenu );
  connect( this, SIGNAL(customContextMenuRequested(const QPoint&)),
        SLOT(showMenu()) );

  setMouseTracking( true );
  objects.resize( 150 );
  selectedObj = -1;

  m_Mode = MEMMODE_ROM;

  updateFontMetrics();

  scrollBar->setRange( 0, 0x10000 / 16 - 9 );
  connect( scrollBar, SIGNAL(valueChanged(int)), SLOT(update()) );

  setMinimumSize( 200, cyLine * 10 );
  scrollTo(0);
  show();
}

bool MemoryView0::event( QEvent *ev ) {
  if ( ev->type() == QEvent::FontChange ) updateFontMetrics();
  return QWidget::event( ev );
}

void MemoryView0::showMenu() {
  QMenu *menu = new QMenu( this );

  QVector <QAction*> actions;
  actions.append( menu->addAction( "RAM plane 0", this, SLOT(showRAM0()) ) );
  actions.append( menu->addAction( "RAM plane 1", this, SLOT(showRAM1()) ) );
  actions.append( menu->addAction( "RAM plane 2", this, SLOT(showRAM2()) ) );
  actions.append( menu->addAction( "ROM", this, SLOT(showROM()) ) );
  actions.append( menu->addAction( "CPU memory", this, SLOT(showCPU()) ) );
  actions.append( menu->addAction( "PPU memory", this, SLOT(showPPU()) ) );

  QActionGroup *group = new QActionGroup( menu );
  for ( int i = 0; i < actions.size(); i++ ) {
	group->addAction( actions[i] );
  	actions[i]->setCheckable( true );
	actions[i]->setChecked( m_Mode == i );
  }

  menu->addSeparator();
  menu->addAction( "Go to address...", this, SLOT(selectAddress()) );

  menu->popup( QCursor::pos() );
}

void MemoryView0::selectAddress() {
  int address = scrollBar->value() * 16;
  bool ok;

  QString text = QInputDialog::getText( this, "Go to address...", "Address:",
	QLineEdit::Normal, QString::number( address, 8 ).rightJustified(6, '0'), &ok );

  if ( ok && !text.isEmpty() ) {
	address = text.toInt( &ok, 8 );
	if ( ok && address > 0 && address < scrollBar->maximum() * 16 )
		scrollTo( address );
  }
}

void MemoryView0::updateFontMetrics() {
  QFontMetrics fm( font() );
  cxChar = fm.averageCharWidth();
  cyLine = fm.height();
  ascent = fm.ascent();

  m_cyLineMemory = cyLine;
}

void MemoryView0::leaveEvent( QEvent* /*ev*/ ) {
  if ( selectedObj != -1 ) {
        selectedObj = -1;
        repaint();
  }
}

void MemoryView0::mouseMoveEvent( QMouseEvent *ev ) {
  QWidget::mouseMoveEvent( ev );

  bool changed = false;
  if ( selectedObj != -1 ) {
        if ( objects[selectedObj].contains( ev->pos() ) ) return;

        selectedObj = -1;
        changed = true;
  }

  for ( int i = 0; i < objects.size(); i++ )
        if ( !objects[i].isEmpty() && objects[i].contains( ev->pos() ) ) {
                selectedObj = i;
                changed = true;
                break;
        }

  if ( changed ) repaint();
}

void MemoryView0::mousePressEvent( QMouseEvent *ev ) {
  QWidget::mousePressEvent( ev );
  if ( ev->button() != Qt::LeftButton ) return;

  if ( selectedObj != -1 && objects[selectedObj].slot )
        QMetaObject::invokeMethod( this, objects[selectedObj].slot,
                Q_ARG(QVariant, objects[selectedObj].parameter),
                Q_ARG(QString, "0") );
}

void MemoryView0::DrawOctalValue( QPainter& p, int x, int y, WORD value ) {
    if ( objIdx < objects.size() ) {
        objects[objIdx] = DebugObject( x, y - ascent, cxChar * 6, cyLine,
                QString::number(value, 8).rightJustified(6, '0'),
                objects[objIdx].slot, objects[objIdx].parameter );
        objIdx++;
    }
    ::DrawOctalValue( p, x, y, value );
}

void MemoryView0::setChangeHandler( const char *slot, QVariant p ) {
    if ( objIdx < objects.size() ) {
        objects[objIdx].slot = slot;
        objects[objIdx].parameter = p;
    }
}

void MemoryView0::setMemoryValue( QVariant p, QString newValue ) {
  bool ok, okHalt;
  WORD address = p.toInt();

  WORD value = newValue.toInt( &ok, 8 );
  if ( ok ) {
        switch (m_Mode) {
            case MEMMODE_RAM0:
            case MEMMODE_RAM1:
            case MEMMODE_RAM2:
                g_pBoard->SetRAMWord(m_Mode, address, value);
                break;
            case MEMMODE_ROM:
                break;
            case MEMMODE_CPU:
                okHalt = g_pBoard->GetCPU()->IsHaltMode();
                g_pBoard->GetCPUMemoryController()->SetWord(address, okHalt, value);
                break;
            case MEMMODE_PPU:
                okHalt = g_pBoard->GetPPU()->IsHaltMode();
                g_pBoard->GetPPUMemoryController()->SetWord(address, okHalt, value);
                break;
        }

        repaint();
  }
}

void MemoryView0::paintEvent( QPaintEvent* /* e */ ) {
  QPainter hdc( this );
  objIdx = 0;

  QColor colorText = hdc.pen().color();
  m_cyLineMemory = cyLine;

    const TCHAR* ADDRESS_LINE = _T(" address  0      2      4      6      10     12     14     16");
    hdc.drawText( 0, ascent, ADDRESS_LINE );

    WORD address = scrollBar->value() * 16;
    int y = 1 * cyLine + ascent;
    for (;;) {  // Draw lines
        ::DrawOctalValue(hdc, 2 * cxChar, y, address);

        int x = 10 * cxChar;
        QString wchars;
        for (int j = 0; j < 8; j++) {  // Draw words as octal value
            // Get word from memory
            WORD word = 0;
            BOOL okValid = TRUE;
            BOOL okHalt = FALSE;
            WORD wChanged = 0;
            switch (m_Mode) {
                case MEMMODE_RAM0:
                case MEMMODE_RAM1:
                case MEMMODE_RAM2:
                    word = g_pBoard->GetRAMWord(m_Mode, address);
                    wChanged = Emulator_GetChangeRamStatus(m_Mode, address);
                    break;
                case MEMMODE_ROM:  // ROM - only 32 Kbytes
                    if (address < 0100000)
                        okValid = FALSE;
                    else
                        word = g_pBoard->GetROMWord(address - 0100000);
                    break;
                case MEMMODE_CPU:
                    okHalt = g_pBoard->GetCPU()->IsHaltMode();
                    word = g_pBoard->GetCPUMemoryController()->GetWordView(address, okHalt, FALSE, &okValid);
                    wChanged = Emulator_GetChangeRamStatus(ADDRTYPE_RAM12, address);
                    break;
                case MEMMODE_PPU:
                    okHalt = g_pBoard->GetPPU()->IsHaltMode();
                    word = g_pBoard->GetPPUMemoryController()->GetWordView(address, okHalt, FALSE, &okValid);
                    if (address < 0100000)
                        wChanged = Emulator_GetChangeRamStatus(ADDRTYPE_RAM0, address);
                    else
                        wChanged = 0;
                    break;
                    break;
            }

            if (okValid)
            {
                hdc.setPen( (wChanged != 0) ? QColor(255,0,0) : colorText );
		setChangeHandler( "setMemoryValue", address );
                DrawOctalValue(hdc, x, y, word);
            }

            // Prepare characters to draw at right
            BYTE ch1 = LOBYTE(word);
            unsigned short wch1 = Translate_KOI8R_Unicode(ch1);
            if (ch1 < 32) wch1 = _T('.');
            wchars += QChar( wch1 );

            BYTE ch2 = HIBYTE(word);
            unsigned short wch2 = Translate_KOI8R_Unicode(ch2);
            if (ch2 < 32) wch2 = _T('.');
            wchars += QChar( wch2 );

            address += 2;
            x += 7 * cxChar;
        }
        hdc.setPen( colorText );

        // Draw characters at right
        int xch = x + cxChar;
        hdc.drawText( xch, y, wchars );

        y += cyLine;
        if (y > height()) break;
    }

    for ( int i = objIdx; i < objects.size(); i++ )
        objects[i] = DebugObject();

    if ( selectedObj != -1 ) {
        hdc.setBackground( QColor(0xff, 0xff, 0xe0) );
        hdc.setBackgroundMode( Qt::OpaqueMode );
        hdc.drawText( objects[selectedObj].x(), objects[selectedObj].y() + ascent, objects[selectedObj].text );
    }
}

LPCTSTR MemoryView0::getMemoryModeName() {
    switch (m_Mode) {
        case MEMMODE_RAM0:  return _T("RAM0");
        case MEMMODE_RAM1:  return _T("RAM1");
        case MEMMODE_RAM2:  return _T("RAM2");
        case MEMMODE_ROM:   return _T("ROM");
        case MEMMODE_CPU:   return _T("CPU");
        case MEMMODE_PPU:   return _T("PPU");
        default:
                            return _T("UKWN");  // Unknown mode
    }
}

// Scroll window to given base address
void MemoryView0::scrollTo(WORD wAddress) {
    scrollBar->setSliderPosition( wAddress / 16 );
}

